/*

              **************************************************
              ****          VEGA - Quanta DCD V1.1          ****
              ****            Trajectory analyzer           ****
              **** Copyright 1996-2000, Alessandro Pedretti ****
              **************************************************


 VEGA writes the binary plot with the endian of the running system.
*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "globdef.h"
#include "globvar.h"
#include "formats.h"

#ifdef WIN32
#include <conio.h>
#endif

#define  RIGA_LEN               256


/**** DCD analyzer ****/

BOOL DCDAnalyzer(char *FileName, ATOMO *Atm, ULONG MolAtm, TRJPAR *Par)
{
  char              *OutMode, Riga[RIGA_LEN];
  DCDHDR            Hdr;
  FILE              *FH, *FH2;
  float             Ene, EneMin, EneMax, MaxRad;
  float             *VdwRad;
  LONG              YPos;
  ULONG             k, Nu, TotAtm, Tm, TmMin;
  ULONG             FrmMin, FrmMax;
  ATOMO             Frm[MAXTRJATM];
  XYZ               BoxMin, BoxMax;
  XYZ               *SfUn;

  BOOL              Ret    = TRUE;
  double            Media  = 0.0;
  double            Media2 = 0.0;

#ifdef VLOGP
  float             *Contrib;
#endif

  /**** Initializations ****/

  switch (Par -> Com) {

  case TRJMS_SURFACE:
  case TRJMS_SURFDIA:
    if (VdwRad = AssignVdwRad(Atm, MolAtm, Prefs.SAS_PROBERAD, &MaxRad)) {
      if (SfUn = CalcSfUn(Prefs.SAS_POINTS, MaxRad, &Nu)) {
        if (!GLOBSW_STDOUT) {
          Ene = CalcSrfArea(Atm, VdwRad, SfUn, Nu, Par, MaxRad);
          if (Par -> Com == TRJMS_SURFACE)
            printf("  Starting area      : %.1f " C_ANGSTROM C_QUAD "\n", Ene);
          else
            printf("  Starting surf. dia.: %.1f " C_ANGSTROM "\n", SQR(Ene / PI));
          }
      } else {
        FREE(VdwRad);
        Ret = FALSE;
      }
    } else Ret = FALSE;
    break;

  case TRJMS_VLOGP:
  #ifdef VLOGP
    if (VdwRad = AssignVdwRad(Atm, MolAtm, VLOGP_PROBE_RAD, &MaxRad)) {
      if (SfUn = CalcSfUn(VLOGP_DOT_DENSITY, MaxRad, &Nu)) {
        if (Contrib = (float *)Alloca(MolAtm * sizeof(float))) {
          FindRing(Atm);
          PrintProg("* Assignig atom types");
          if (Ret = LogPCalc(Atm, MolAtm, &Ene, &EneMin, "BROTO", Contrib)) {
            if (!GLOBSW_STDOUT)
              printf("  Starting vlogP     : %.4f\n",
                     CalcVLogP(Atm, VdwRad, MaxRad, SfUn, Nu, Contrib));

          } else {
            FREE(VdwRad);
            FREE(SfUn);
            FREE(Contrib);
          }
        } else {
          FREE(VdwRad);
          FREE(SfUn);
          Ret = FALSE;
        }
      } else {
        FREE(VdwRad);
        Ret = FALSE;
      }
    } else Ret = FALSE;
#else
    Ret = PrintVlogPErr();
#endif
    break;

  case TRJMS_VOLUME:
  case TRJMS_VOLDIA:
      if (VdwRad = AssignVdwRad(Atm, MolAtm, 0.0, &MaxRad)) {
        CalcBox(Atm, VdwRad, &BoxMin, &BoxMax);
        if (!GLOBSW_STDOUT) {
          Ene =  CalcVolume(Atm, VdwRad, Prefs.VOL_DEN, &BoxMin, &BoxMax);
          if (Par -> Com == TRJMS_VOLUME)
            printf("  Starting volume    : %.1f " C_ANGSTROM C_CUBE "\n", Ene);
          else
            printf("  Starting vol. dia. : %.1f " C_ANGSTROM "\n",
                   pow(6.0 * Ene / PI, 1.0 / 3.0));
        }
      } else Ret = FALSE;
    break;

  case TRJMS_DIPOLE:
    if (!ChkCharges(Atm, MolAtm, NULL))
      Ret = Errore("Unassigned atomic charges\n" \
                   "You need to fix it with -c[TEMPLATE] option");
    break;
  }

  if (Ret) {

    /**** Set the write mode ****/

    switch (Par -> OutForm -> Id) {
    case FORM_PLOT_CSV:
    case FORM_PLOT_QUANTA:
      OutMode = "w";
      break;
    case FORM_PLOT_BIN:
      OutMode = "wb";
      break;
    default :
      Ret = Errore("Unsupported output format for trajectory");
    }

    if ((Ret) && (Ret = AtmTrans(Atm, MolAtm, Par))) {
      if (FH = fopen(Prefs.DCD_FILE, "rb")) {
        if (Ret = ForRead(FH, &Hdr, sizeof(DCDHDR))) {
  #ifdef BIG_ENDIAN
          Swap(&Hdr.TotFrames);
          Swap(&Hdr.StartTime);
          Swap(&Hdr.FrameStep);
          Swap(&Hdr.NFixed);
  #endif
          if (Ret = ForSeek(FH,1)){
            if (Ret = ForRead(FH, &TotAtm, sizeof(ULONG))) {
  #ifdef BIG_ENDIAN
              Swap(&TotAtm);
  #endif
              if (TotAtm == MolAtm) {

                /**** Calculates the real number of frames ****/

                k = ftell(FH);
                fseek(FH, 0, SEEK_END);
                Hdr.TotFrames = (ftell(FH) - k) / (sizeof(float) * 3 * TotAtm +
                                 sizeof(ULONG) * 6);
                fseek(FH, k, SEEK_SET);

                if (!GlobSwitch[4])
                  printf("  Total frames       : %d\n" \
                         "  Frame step         : %d fs\n" \
                         "  Simulation time    : %d ps\n" \
                         "  Total atoms        : %d\n",
                         Hdr.TotFrames, Hdr.FrameStep,
                         (Hdr.TotFrames * Hdr.FrameStep) / 1000,
                         TotAtm);

                /**** Read .ENE file (if exists) ****/

                ChangeExt(FileName, "ENE");
                if (FH2 = fopen(FileName, "r")) {
                  k = 1;
                  while(fgets(Riga, RIGA_LEN, FH2)) {
                    sscanf(Riga, "%d %*s %*s %*s %f", &Tm, &Ene);
                    if (k == 1) {
                      TmMin  = Tm;
                      EneMin = Ene;
                    } else if (Ene < EneMin) {
                      EneMin = Ene;
                      TmMin  = Tm;
                    }
                    ++k;
                    if ((!fgets(Riga, RIGA_LEN, FH2)) ||
                        (!fgets(Riga, RIGA_LEN, FH2))) break;
                  }
                  if (!GlobSwitch[4])
                    printf("  Lowest energy frame: %d (%.4f Kcal/mol)\n",
                           TmMin / Hdr.FrameStep, EneMin);
                  fclose(FH2);
                }

                /**** Make calculations ****/

                if (*Par -> OutFile) {
                  PrintProg("* Saving plot");
                  if (!strchr(Par -> OutFile, '.')) {
                    strcat(Par -> OutFile, ".");
                    strcat(Par -> OutFile, Par -> OutForm -> Ext);
                  }
                  FH2 = fopen(Par -> OutFile, OutMode);
#ifdef WIN32
                  YPos = wherey();
#else
                  YPos = 0;
#endif
                } else {
                  FH2  = stdout;
                  YPos = -1;
                }
                if (FH2) {

                  for(k = 0; (Ret) && (k < Hdr.TotFrames); ++k) {
                    switch(Par -> Com) {
                    case TRJMS_DISTANCE:
                      if (Ret = DCDRead(FH, TotAtm, Par, (ATOMO *)&Frm))
                        Ene = SQR(Quad(Frm[0].x - Frm[1].x) + Quad(Frm[0].y - Frm[1].y) +
                                  Quad(Frm[0].z - Frm[1].z));
                      break;

                    case TRJMS_ANGLE:
                      if (Ret = DCDRead(FH, TotAtm, Par, (ATOMO *)&Frm))
                        Ene = (float)BondAngle(&Frm[0], &Frm[1], &Frm[2]);
                      break;

                    case TRJMS_TORSION:
                      if (Ret = DCDRead(FH, TotAtm, Par, (ATOMO *)&Frm))
                        Ene = (float)Torsion(&Frm[0], &Frm[1], &Frm[2], &Frm[3]);
                      break;

                    case TRJMS_PLANEANG:
                      if (Ret = DCDRead(FH, TotAtm, Par, (ATOMO *)&Frm))
                        Ene = (float)PlaneAngAtm((ATOMO *)&Frm) * RAD_TO_DEG;
                      break;
#ifdef VLOGP
                    case TRJMS_VLOGP:
                      if (Ret = DCDReadFrm(FH, Atm)) {
                        Ene = (float)CalcVLogP(Atm, VdwRad, MaxRad, SfUn, Nu,
                                               Contrib);
                      }
                      break;
#endif
                    case TRJMS_VOLUME:
                    case TRJMS_VOLDIA:
                      if (Ret = DCDReadFrm(FH, Atm)) {
                        CalcBox(Atm, VdwRad, &BoxMin, &BoxMax);
                        Ene = (float)CalcVolume(Atm, VdwRad, Prefs.VOL_DEN, &BoxMin, &BoxMax);
                        if (Par -> Com == TRJMS_VOLDIA)
                          Ene = pow(6.0 * Ene / PI, 1.0 / 3.0);
                      }
                      break;

                    case TRJMS_DIPOLE:
                      if (Ret = DCDReadFrm(FH, Atm))
                        Ene = (float)CalcDipole(Atm);
                      break;
                    case TRJMS_SURFACE:
                    case TRJMS_SURFDIA:
                      if (Ret = DCDReadFrm(FH, Atm)) {
                        Ene = (float)CalcSrfArea(Atm, VdwRad, SfUn, Nu, Par, MaxRad);
                        if (Par -> Com == TRJMS_SURFDIA) Ene = SQR(Ene / PI);
                      }
                      break;
                    }

                    if (Ret) {
                      Media  += Ene;
                      Media2 += Quad(Ene);
		      if (!k) {
		        EneMin = Ene;
			EneMax = Ene;
			FrmMin = 1;
			FrmMax = 1;
		      } else {
		        if (Ene < EneMin) {
			  EneMin = Ene;
			  FrmMin = k + 1;
			}
			if (Ene > EneMax) {
			  EneMax = Ene;
			  FrmMax = k + 1;
			}
		      }
                      if (!GlobSwitch[4]) {
#ifdef WIN32
                        gotoxy(1, YPos);
                        clreol();
                        cprintf("  Processing frame %d/%d (%.4f)", k + 1, Hdr.TotFrames, Ene);
#else
                        printf("  Processing frame %d/%d (%.4f)          \n\033M",
                               k + 1, Hdr.TotFrames, Ene);
#endif
                      }

                      /**** Write data ****/

                      ++k;
                      switch(Par -> OutForm -> Id) {
                      case FORM_PLOT_CSV:           /* Comma separted values */
                        if (fprintf(FH2, "%5d; %14.6f\n", k, Ene) < 0)
                          Ret = PrintDosErr();
                        break;

                      case FORM_PLOT_QUANTA:         /* Quanta plot          */
                        if (fprintf(FH2, "%5d.000000,%14.6f\n", k, Ene) < 0)
                          Ret = PrintDosErr();
                        break;

                      case FORM_PLOT_BIN:            /* Binary Plot          */
                        if ((fwrite(&k  , sizeof(ULONG), 1, FH2) != 1) ||
                            (fwrite(&Ene, sizeof(float), 1, FH2) != 1))
                          Ret = PrintDosErr();
                        break;
                      }
                      --k;
                    }
                  } /* End of frame loop */
                  if (!GlobSwitch[4]) {
#ifdef WIN32
                    gotoxy(1, YPos);
                    clreol();
#else
                    printf("%40s\n\033M", " ");
#endif

                    printf("  Range              : %.4f (%d) <-> %.4f (%d)\n" \
                           "  Average value      : %.4f " C_PLUSMIN "%.4f\n",
			   EneMin, FrmMin, EneMax, FrmMax,
                           Media / Hdr.TotFrames,
                           SQR((Media2 - (Quad(Media) / Hdr.TotFrames)) / (Hdr.TotFrames - 1)));
                  }
                  if (FH2 != stdout) fclose(FH2);
                } else Ret = PrintDosErr();
              } else Ret = Errore("Uncompatible trajectory file");
            }
          }
        }
        fclose(FH);
      } else Ret = PrintDosErr();
    }
    switch(Par -> Com) {
    case TRJMS_SURFACE:
    case TRJMS_SURFDIA:
      FREE(VdwRad);
      FREE(SfUn);
      break;

#ifdef VLOGP
    case TRJMS_VLOGP:
      FREE(VdwRad);
      FREE(SfUn);
      FREE(Contrib);
      break;
#endif

    case TRJMS_VOLUME:
    case TRJMS_VOLDIA:
      FREE(VdwRad);
      break;
    }
  }

  return Ret;
}


/**** Read specified atoms of a trajectory frame ****/

BOOL DCDRead(FILE *FH, ULONG TotAtm, TRJPAR *Par, ATOMO *Frm)
{
  ULONG         AttPos;
  UWORD         k;

  BOOL          Ret     = TRUE;
  ULONG         PrecPos = 0;

  /**** Read X coordinates ****/

  fseek(FH, sizeof(ULONG), SEEK_CUR);
  for(k = 0; (Ret) && (k < Par -> Num); ++k) {
    AttPos = (Par -> Atm[k].Num - 1) * sizeof(float);
    fseek(FH, AttPos - PrecPos, SEEK_CUR);
    if (fread(&Frm[k].x, sizeof(float), 1, FH) == 1) {
#ifdef BIG_ENDIAN
        Swap(&Frm[k].x);
#endif
      PrecPos = AttPos + sizeof(float);
    } else Ret = PrintDosErr();
  }

  /**** Read Y coordinates ****/

  if (Ret) {
    fseek(FH, (TotAtm - Par -> Atm[k - 1].Num) * sizeof(float) +
              sizeof(ULONG) * 2, SEEK_CUR);
    PrecPos = 0;
    for(k = 0; (Ret) && (k < Par -> Num); ++k) {
      AttPos = (Par -> Atm[k].Num - 1) * sizeof(float);
      fseek(FH, AttPos - PrecPos, SEEK_CUR);
      if (fread(&Frm[k].y, sizeof(float), 1, FH) == 1) {
#ifdef BIG_ENDIAN
        Swap(&Frm[k].y);
#endif
        PrecPos = AttPos + sizeof(float);
      } else Ret = PrintDosErr();
    }

    /**** Read Z coordinates ****/

    if (Ret) {
      fseek(FH, (TotAtm - Par -> Atm[k - 1].Num) * sizeof(float) +
            sizeof(ULONG) * 2, SEEK_CUR);
      PrecPos = 0;
      for(k = 0; (Ret) && (k < Par -> Num); ++k) {
        AttPos = (Par -> Atm[k].Num - 1) * sizeof(float);
        fseek(FH, AttPos - PrecPos, SEEK_CUR);
        if (fread(&Frm[k].z, sizeof(float), 1, FH) == 1) {
#ifdef BIG_ENDIAN
        Swap(&Frm[k].z);
#endif
          PrecPos = AttPos + sizeof(float);
        } else Ret = PrintDosErr();
      }
      fseek(FH, (TotAtm - Par -> Atm[k - 1].Num) * sizeof(float) +
            sizeof(ULONG), SEEK_CUR);
    }
  }

  return Ret;
}


/**** Read all atoms of a trajectory frame ****/

BOOL DCDReadFrm(FILE *FH, ATOMO *InizAtm)
{
  register ATOMO        *Atm;

  BOOL                  Ret = TRUE;

  /**** Read X coordinates ****/

  fseek(FH, sizeof(ULONG), SEEK_CUR);
  for(Atm = InizAtm;(Ret) && (Atm); Atm = Atm -> Ptr) {
    if (fread(&Atm -> x, sizeof(float), 1, FH) != 1)
      Ret = PrintDosErr();
#ifdef BIG_ENDIAN
    else Swap(&Atm -> x);
#endif
  } /* End of atom loop */

  if (Ret) {

    /**** Read Y coordinates ****/

    fseek(FH, sizeof(ULONG) * 2, SEEK_CUR);
    for(Atm = InizAtm;(Ret) && (Atm); Atm = Atm -> Ptr) {
      if (fread(&Atm -> y, sizeof(float), 1, FH) != 1)
        Ret = PrintDosErr();
#ifdef BIG_ENDIAN
      else Swap(&Atm -> y);
#endif
    }
    if (Ret) {

      /**** Read Z coordinates ****/

      fseek(FH, sizeof(ULONG) * 2, SEEK_CUR);
      for(Atm = InizAtm;(Ret) && (Atm); Atm = Atm -> Ptr) {
        if (fread(&Atm -> z, sizeof(float), 1, FH) != 1)
          Ret = PrintDosErr();
#ifdef BIG_ENDIAN
        else Swap(&Atm -> z);
#endif
      }
      if (Ret) fseek(FH, sizeof(ULONG), SEEK_CUR);
    }
  }

  return Ret;
}
