/*

             **************************************************
             ****             VEGA - Utils V1.1            ****
             ****          Utilities & Subroutines         ****
             **** Copyright 1996-2000, Alessandro Pedretti ****
             **************************************************


*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>


#include "globdef.h"
#include "globvar.h"

#ifndef WIN32
#include <unistd.h>
#include <errno.h>
#else
#include <stdlib.h>
#endif

#ifdef AMIGA
#include <dos/dos.h>
#include <exec/types.h>
#include <exec/memory.h>
#include <exec/execbase.h>
#endif

extern int   errno;

static ULONG AtmLoaded = 0;


/**** Add a value to the radius matrix ****/

void AddToRad(register ATOMO *Atm, register float *Rad, register float Val)
{
  while(Atm) {
    *Rad += Val;
    Atm = Atm -> Ptr;
    ++Rad;
  }
}


/**** Allocate the memory with error check ****/

void *Alloca(ULONG Mem)
{
  void		*Ind;

  if ((Ind = (void *)ALLOC(Mem)) == NULL)
    Errore("Out of memory");

  return Ind;
}


/**** Allocates the description structure for one atom ****/

ATOMO *AllocAtm(ATOMO **InizAtm, ULONG *TotAtomi)
{
  register ATOMO   *Atm;
  static ATOMO     *PrecAtm;

  if (Atm = (ATOMO *)Alloca(sizeof(ATOMO))) {
    if (!*TotAtomi) *InizAtm = Atm;
    else PrecAtm -> Ptr = Atm;
    ++*TotAtomi;
    PrecAtm           = Atm;
    *Atm  -> ResSeq.C = '1';
    Atm  -> ChainID   = ' ';
    Atm  -> Num       = ++AtmLoaded;
    *Atm -> Elem.C    = '?';
    *Atm -> Pot.C     = '?';
    Atm -> ResName.L = *(LONG *)"UNK\0";
  } else if (*TotAtomi) {
    PDBFree(*InizAtm);
    *InizAtm = NULL;
  }

  return Atm;
}


/**** Mesures the bond angle ****/

double BondAngle(ATOMO *a, ATOMO *b, ATOMO *c)
{
  double  Angle;
  double  Dist;
  double  Cos_Theta;

  Dist = (double)Distance(a,b) * (double)Distance(b,c);
  Cos_Theta = ((a -> x - b -> x) * (c -> x - b -> x) + (a -> y - b -> y) * (c -> y - b -> y) +
         (a -> z - b -> z) * (c -> z - b -> z)) / Dist;
  if ((Cos_Theta  + (double)1.0) < (double)0.0001)
    Angle = (double)180.0;
  else Angle = (acos(Cos_Theta)) * RAD_TO_DEG;

  return Angle;
}


/**** Change or add the file extension ****/

void ChangeExt(char *Str, char *Ext)
{
  register char *Ptr;

  for(Ptr = Str + strlen(Str) - 1; Ptr != Str; --Ptr)
    if (*Ptr == '.') {
      strcpy(Ptr + 1, Ext);
      break;
    }
  if ((Ptr == Str) && (*Ptr != '.')) {
    strcat(Ptr, ".");
    strcat(Ptr, Ext);
  }
}

/**** Check if the charges are assigned ****/

BOOL ChkCharges(register ATOMO *Atm, ULONG TotAtm, float *TotCharge)
{
  BOOL            Ret       = TRUE;
  register float  Chg       = 0;
  register ULONG  Uncharged = 0;

  do {
    if (Atm -> Charge == 0.0) ++Uncharged;
    Chg += Atm -> Charge;
    Atm = Atm -> Ptr;
  } while(Atm);
  if (Uncharged != TotAtm) {
    if (TotCharge) *TotCharge = Chg;
  } else Ret = FALSE;

  return Ret;
}


/**** Check if the force field is assigned ****/

BOOL ChkForceField(register ATOMO *Atm, ULONG TotAtm)
{
  register ULONG  Unassigned = 0;

  do {
    if (Atm -> Pot.L == 0x3f000000) ++Unassigned;
    Atm = Atm -> Ptr;
  } while(Atm);
  if (Unassigned == TotAtm) return FALSE;
  else return TRUE;
}


/**** Counts residues ****/

ULONG CountRes(register ATOMO *Atm)
{
  register ULONG  PrecResName = 0;
  register ULONG  PrecResSeq  = 0;
  register ULONG  TotRes      = 0;

  while(Atm) {
    if ((Atm -> ResName.L != (LONG)PrecResName) ||
        (Atm -> ResSeq.L  != (LONG)PrecResSeq)) {
      PrecResName = Atm -> ResName.L;
      PrecResSeq  = Atm -> ResSeq.L;
      ++TotRes;
    }
    Atm = Atm -> Ptr;
  }

  return TotRes;
}


/**** Vectorial cross product ****/

void CrossProd(VECTYPE *vect1, VECTYPE *vect2, VECTYPE *normal)
{
  normal -> x = vect1 -> y * vect2 -> z - vect2 -> y * vect1 -> z;
  normal -> y = vect2 -> x * vect1 -> z - vect1 -> x * vect2 -> z;
  normal -> z = vect1 -> x * vect2 -> y - vect2 -> x * vect1 -> y;
}


/**** Determinat of 2x2 matrix ****/

float Det2x2(float A, float B, float C, float D)
{
  return (A * D - B * C);
}


/**** Dot product ****/

double Dot(VECTYPE *vect1, VECTYPE *vect2)
{
  double  Dot_prod;

  Dot_prod = vect1 -> x * vect2 -> x + vect1 -> y * vect2 -> y + vect1 -> z *
  vect2 -> z;
  return Dot_prod;
}


/**** Shows an error ****/

BOOL Errore(const char *Err, ...)
{
  char          Buf[128];
  va_list       vl;

#ifndef WIN32
  strcpy(Buf, "\n\033[33mERROR:\033[0m ");
#else
  strcpy(Buf, "\nERROR: ");
#endif
  va_start(vl, Err);
  strcat(Buf, Err);
  strcat(Buf, ".\n\n");
  vfprintf(stderr, Buf, vl);
  va_end(vl);

  return FALSE;
}


/**** Find a residue ****/

ULONG FindResPtr(register ATOMO *Atm, ATOMO **ResPtr, LONG NomeRes, LONG NRes)
{
  register UWORD  Chk;
  register ULONG  NAtm = 0;

  *ResPtr = 0;

  while(Atm) {
    Chk = 0;
    if (!NomeRes) Chk = 1;
    else if (Atm -> ResName.L == NomeRes) Chk = 1;
    if (!NRes) ++Chk;
    else if (Atm -> ResSeq.L == NRes) ++Chk;
    if (Chk == 2) {
      if (!*ResPtr) *ResPtr = Atm;
      ++NAtm;
    }
    Atm = Atm -> Ptr;
  }

  return NAtm;
}


/**** Reads a record in Fortran format ****/

BOOL ForRead(FILE *FH, void *Ptr, ULONG Len)
{
  BOOL            Ret = TRUE;

  fseek(FH, sizeof(int), SEEK_CUR);
  if (fread(Ptr, Len, 1, FH) == 1)
    fseek(FH, sizeof(int), SEEK_CUR);
  else Ret = PrintDosErr();

  return Ret;
}


/**** Performs Seek from/to Fortran records ****/

BOOL ForSeek(FILE *FH, ULONG Rec)
{
  unsigned int  Len;
  BOOL    Ret = TRUE;

  do {
    if (fread(&Len, sizeof(int), 1, FH)) {
#ifdef BIG_ENDIAN
      Swap(&Len);
#endif
      fseek(FH, Len + sizeof(int), SEEK_CUR);
    } else Ret = PrintDosErr();
    --Rec;
  } while ((Rec) && (Ret));

  return Ret;
}


/**** Copy a string from Fortran to C ****/

void ForStrCpy(register char *Dest, register char *Src, ULONG Len)
{
  register ULONG  k;

  for(k = 0;(k < Len) && (Src[k] != ' ');++k)
    Dest[k] = Src[k];
}


/**** Writes a record in Fortran unformatted format ****/

BOOL ForWrite(FILE *FH, void *Num, ULONG Size)
{
  BOOL      Ret = TRUE;

#ifdef BIG_ENDIAN
  ULONG     Sz  = SwapI(Size);
#else
  ULONG     Sz  = Size;
#endif

  if ((fwrite(&Sz, sizeof(ULONG), 1, FH) != 1) ||
      (fwrite(Num, Size, 1, FH) != 1) ||
      (fwrite(&Sz, sizeof(ULONG), 1, FH) != 1))
    Ret = PrintDosErr();

  return Ret;
}


/**** Copy in a string the path of Vega ****/

BOOL GetPrgPath(char *Str)
{
  char    *CharPtr;

  if (!getenv("VEGADIR")) {
    Errore("Please set the environment variable `VEGADIR'");
    return FALSE;
  }
  strcpy(Str, (char *)getenv("VEGADIR"));
  CharPtr = Str + strlen(Str) - 1;

#ifdef WIN32
  if (*CharPtr != '\\') strcat(Str, "\\");
  strcat(Str, "Data\\");
#else
  if ((*CharPtr != ':') && (*CharPtr != '/')) strcat(Str, "/");
  strcat(Str, "Data/");
#endif

  return TRUE;
}



double Magnitude(VECTYPE *vect1)
{
  double  Mag;

  Mag = SQR(Quad(vect1 -> x) + Quad(vect1 -> y) + Quad(vect1 -> z));

  return Mag;
}


/**** Free a memory list of atoms ****/

void PDBFree(register ATOMO *Atm)
{
  register ATOMO	*Succ;

  if (Atm) {
    do {
      Succ = Atm -> Ptr;
      FREE(Atm);
      Atm = Succ;
    } while(Atm);
  }
}


/**** Calculate the angle between two planes ****/
/**** starting from plane equations          ****/

float PlaneAngle(PLANE *P0, PLANE *P1)
{
   return acos((P0 -> a * P1 -> a + P0 -> b * P1 -> b + P0 -> c * P1 -> c) /
               (SQR(Quad(P0 -> a) + Quad(P0 -> b) + Quad(P0 -> c)) * \
                SQR(Quad(P1 -> a) + Quad(P1 -> b) + Quad(P1 -> c))));
}


/**** Calculate the angle between two planes ****/
/**** starting from atomic coordinates        ****/

float PlaneAngAtm(ATOMO *Atm)
{
  PLANE         P0, P1;

  PlaneCof(&P0, Atm, Atm + 1, Atm + 2);
  PlaneCof(&P1, Atm + 3, Atm + 4, Atm + 5);

  return (float)PlaneAngle(&P0, &P1);
}


/**** Calculate the equation coefficients of plene through 3 atoms ****/

void PlaneCof(PLANE *Plane, ATOMO *A0, ATOMO *A1, ATOMO *A2)
{
  Plane -> a = Det2x2(A1 -> y - A0 -> y, A1 -> z - A0 -> z,
                      A2 -> y - A0 -> y, A2 -> z - A0 -> z);
  Plane -> b = Det2x2(A1 -> x - A0 -> x, A1 -> z - A0 -> z,
                      A2 -> x - A0 -> x, A2 -> z - A0 -> z);
  Plane -> c = Det2x2(A1 -> x - A0 -> x, A1 -> y - A0 -> y,
                      A2 -> x - A0 -> x, A2 -> y - A0 -> y);
  Plane -> d = A0 -> y * Plane -> b - A0 -> x * Plane -> a -
               A0 -> z * Plane -> c;
}


void PointToVect(ATOMO *pt, VECTYPE *v)
{
  v -> x = pt -> x;
  v -> y = pt -> y;
  v -> z = pt -> z;
}


/**** Print an error with atom ****/

BOOL PrintAtmErr(char *Err, ATOMO *Atm)
{
  if (!GlobSwitch[4])
    fprintf(stderr, "%s: %-7d %-4.4s %-2.2s %.4s_%.4s\n", Err,
            Atm -> Num, Atm -> Name.C, Atm -> Elem.C, Atm -> ResName.C,
            Atm -> ResSeq.C);

  return FALSE;
}


/**** Show a system error ****/

BOOL PrintDosErr(void)
{
#ifdef WIN32
  fprintf(stderr, "\nDOS ERROR: %s\n", strerror(errno));
#else
  fprintf(stderr, "\n\033[33mDOS ERROR:\033[0m %s.\n\n",
          strerror(errno));
#endif

  return FALSE;
}


/**** Print loading format ****/

void PrintLoad(char *Msg, ULONG NMol)
{
  if (!GlobSwitch[4])
    printf("* Loading file %d as %s\n", NMol, Msg);
}

/**** Print a progress message ****/

void PrintProg(char *Msg)
{
  if (!GlobSwitch[4]) printf("%s\n", Msg);
}


/**** Print warning message ****/

void PrintWarn(char *Msg)
{
  if (!GlobSwitch[4])
#ifdef WIN32
    printf("\nWARNING: %s\n\n", Msg);
#else
    printf("\n\033[33mWARNING:\033[0m %s\n\n", Msg);
#endif
}


/**** Remove a single atom from list ****/

void RemoveAtm(ATOMO **Atm, ATOMO **PrecAtm, ATOMO **InizAtm)
{
  if (*Atm == *InizAtm) {
    *InizAtm = (*Atm) -> Ptr;
    *PrecAtm = (*Atm) -> Ptr;
    FREE(*Atm);
    *Atm = *PrecAtm;
  } else {
    (*PrecAtm) -> Ptr = (*Atm) -> Ptr;
    if (((*Atm) -> Flags > 1) && ((*PrecAtm) -> Flags <= 1))
      (*PrecAtm) -> Flags += (BYTE)2;
    FREE(*Atm);
    *Atm = (*PrecAtm) -> Ptr;
  }
}


/**** Remove hydrogens ****/

void RemoveH(ATOMO **InizAtm, ULONG *TotAtm)
{
  register ATOMO   *Con;
  register UWORD  k, j;

  ATOMO      *PrecAtm = *InizAtm;
  ATOMO      *Atm     = *InizAtm;
  ULONG     Count    = 1;

  PrintProg("* Removing hydrogens");

  while(Atm) {
    if (Atm -> Elem.S == 0x4800) {
      for(Con = *InizAtm;Con;Con = Con -> Ptr) {
        for(k = 0;(k < MAXBOND) && (Con -> Conn[k]); ++k) {
          if (Con -> Conn[k] == Atm) {
            for(j = k; j < (MAXBOND - 1); ++j) {
              Con -> Conn[j]  = Con -> Conn[j + 1];
              Con -> Order[j] = Con -> Order[j + 1];
            }
          }
        }
      }
      --*TotAtm;
      RemoveAtm(&Atm, &PrecAtm, InizAtm);
    } else {
      PrecAtm    = Atm;
      Atm -> Num = Count++;
      Atm        = Atm -> Ptr;
    }
  }
}


/**** Removes water molecules ****/

void RemoveH2O(ATOMO **InizAtm, ULONG *TotAtm)
{
  ATOMO      *Atm     = *InizAtm;
  ATOMO      *PrecAtm = *InizAtm;
  register ULONG  Count    = 1;

  PrintProg("* Removing waters");

  if (RenameWat(*InizAtm)) {
    Atm     = *InizAtm;
    PrecAtm = Atm;
    while(Atm) {
      if ((Atm -> ResName.C[0] ==  'H') && (Atm -> ResName.C[1] ==  'O') &&
          (Atm -> ResName.C[2] ==  'H')) {
        --*TotAtm;
        RemoveAtm(&Atm, &PrecAtm, InizAtm);
      } else {
        PrecAtm    = Atm;
        Atm -> Num = Count++;
        Atm        = Atm -> Ptr;
      }
    }
  }
}


/**** Remove a residue ****/

void RemoveRes(register ATOMO **Atm, ATOMO **PrecAtm, ATOMO **InizAtm,
               register ULONG *TotAtm)
{
  LONG          ResSeq = (*Atm) -> ResSeq.L;

  do {
    RemoveAtm(Atm, PrecAtm, InizAtm);
    --*TotAtm;
  } while((*Atm) && ((*Atm) -> ResSeq.L == ResSeq));
}


/**** Rename a residue ****/

ATOMO *RenameRes(register ATOMO *Atm, LONG Res)
{
  QCHAR             ResName;
  register ATOMO		*PrecAtm = Atm;
  register LONG 	ResSeq   = Atm -> ResSeq.L;

  if ((Atm -> ResName.C[3] == '+') || (Atm -> ResName.C[3] == '-'))
    ((char *)&Res)[2] = Atm -> ResName.C[3];

  ResName.L = Atm -> ResName.L;
  while((Atm) && (ResName.L == Atm -> ResName.L) && (ResSeq == Atm -> ResSeq.L)) {
    Atm -> ResName.L = Res;
    PrecAtm = Atm;
    Atm     = Atm -> Ptr;
  }

  return PrecAtm;
}


/**** Rename waters ****/

BOOL RenameWat(register ATOMO *Atm)
{
  register ATOMO		*H1, *H2;

  static char		*H2O = "HOH";
  BOOL			Ret  = FALSE;

  while(Atm) {
    H1 = Atm -> Conn[0];
    H2 = Atm -> Conn[1];

    if ((Atm -> Elem.S == 0x4f00) && (Atm -> NSost == 2) &&
        (H1 -> Elem.S == 0x4800) && (H2 -> Elem.S == 0x4800)) {
      strncpy(Atm -> ResName.C, H2O, 3);
      strncpy(H1  -> ResName.C, H2O, 3);
      strncpy(H2  -> ResName.C, H2O, 3);
      Ret = TRUE;
    }
    Atm = Atm -> Ptr;
  }

  return Ret;
}


/**** Renumbers residues ****/

void RenRes(register ATOMO *Atm)
{
  char      Tmp[6];

  register ULONG  PrecResSeq = Atm -> ResSeq.L;
  register ULONG  ResSeq     = Prefs.RENSTART;

  PrintProg("* Renumbering residues");

  while(Atm) {
    if (Atm -> ResSeq.L != (LONG)PrecResSeq) {
      PrecResSeq = Atm -> ResSeq.L;
      ++ResSeq;
    }
    sprintf(Tmp, "%d", ResSeq);
    ResConv(&Atm -> ResSeq, Tmp);
    Atm = Atm -> Ptr;
  }
}


/**** Translates a string into ResSeq ****/

void ResConv(QCHAR *Res, register char *Str)
{
  register char   *Dest = (*Res).C;
  register UWORD  k;

  (*Res).L = (ULONG)NULL;
  for(k = 0;(*Str) && (k < 4); ++k) {
   *Dest = *Str;
   ++Str;
   ++Dest;
  }
}


/**** Set the residue type ****/

void SetTypeRes(QCHAR *Res, UBYTE Car)
{
  if (Res -> L != *(LONG *)"ASP-") {
    if ((Res -> C[3] == '+') || (Res -> C[3] == '-')) {
      Res -> C[2] = Res -> C[3];
    }
  } else Res -> L = *(LONG *)"AP-";

  Res -> C[3] = Car;
}


/**** Change endian ****/

void Swap(register void *Val)
{
  register UBYTE    T;

  T = ((UBYTE *)Val)[0];
  ((UBYTE *)Val)[0] =  ((UBYTE *)Val)[3];
  ((UBYTE *)Val)[3] = T;
  T = ((UBYTE *)Val)[1];
  ((UBYTE *)Val)[1] =  ((UBYTE *)Val)[2];
  ((UBYTE *)Val)[2] = T;
}


/**** Change the endian for float ****/

double SwapD(double Num)
{
  register UBYTE    T;
  register UBYTE    *Val = (UBYTE *)&Num;

  T = ((UBYTE *)Val)[0];
  ((UBYTE *)Val)[0] =  ((UBYTE *)Val)[7];
  ((UBYTE *)Val)[7] = T;
  T = ((UBYTE *)Val)[1];
  ((UBYTE *)Val)[1] =  ((UBYTE *)Val)[6];
  ((UBYTE *)Val)[6] = T;
  T = ((UBYTE *)Val)[2];
  ((UBYTE *)Val)[2] =  ((UBYTE *)Val)[5];
  ((UBYTE *)Val)[5] = T;
  T = ((UBYTE *)Val)[3];
  ((UBYTE *)Val)[3] =  ((UBYTE *)Val)[4];
  ((UBYTE *)Val)[4] = T;

  return Num;
}

/**** Change the endian for float ****/

float SwapF(float Val)
{
  Swap(&Val);
  return Val;
}


/**** Change the endian for int ****/

int SwapI(int Val)
{
  Swap(&Val);
  return Val;
}


/**** Change the endian for word ****/

void SwapW(register void *Val)
{
  register UBYTE    T;

  T = ((UBYTE *)Val)[0];
  ((UBYTE *)Val)[0] =  ((UBYTE *)Val)[1];
  ((UBYTE *)Val)[1] = T;
}


/**** Change the endian (if needed) and writes a long ****/

int SwapWrite(FILE *FH, void *Val)
{
  LONG          Num = *(LONG *)Val;
#ifdef BIG_ENDIAN
  Swap(&Num);
#endif

  return (int)fwrite(&Num, sizeof(ULONG), 1, FH);
}


/**** Mesure a torsion angle ****/

double Torsion(ATOMO *a, ATOMO *b, ATOMO *c, ATOMO *d)
{
  VECTYPE b1, b2, b3;
  VECTYPE c1, c2, c3;
  double  l1, l2;
  VECTYPE v1, v2, v3, v4;
  double  angle;

  PointToVect(a, &v1);
  PointToVect(b, &v2);
  PointToVect(c, &v3);
  PointToVect(d, &v4);

  VectDiff(&v1, &v2, &b1);
  VectDiff(&v2, &v3, &b2);
  VectDiff(&v3, &v4, &b3);

  CrossProd(&b1, &b2, &c1);
  CrossProd(&b2, &b3, &c2);
  CrossProd(&c1, &c2, &c3);

  l1 = Magnitude(&c1);
  l2 = Magnitude(&c2);

  if ((l1 * l2) != 0) {
    angle = VectAng(&c1, &c2);
    if (Dot(&b2, &c3) > 0.0) angle = angle * -1.0;
  } else angle = 0;

  return angle;
}


/**** Change a string to uppercase ****/

void Upper(register char *Str)
{
  do *Str = (char)toupper(*Str); while(*++Str);
}


double VectAng(VECTYPE *vect1, VECTYPE *vect2)
{
  double  angle;
  double  mag;
  double  dp;
  double  rad_ang;

  mag = Magnitude(vect1) * Magnitude(vect2);
  dp  = Dot(vect1, vect2) / mag;
  if (dp < -0.999999) dp = -0.9999999;
  if (dp > 0.9999999) dp = 0.9999999;
  rad_ang = acos(dp);
  angle   = RAD_TO_DEG * rad_ang;

  return angle;
}


void VectDiff(VECTYPE *vect1, VECTYPE *vect2, VECTYPE *vect_sm)
{
  vect_sm -> x = vect1 -> x - vect2 -> x;
  vect_sm -> y = vect1 -> y - vect2 -> y;
  vect_sm -> z = vect1 -> z - vect2 -> z;
}


/**** Input prompt for a string ****/

void VegaGetString(char *Prompt, char *Str, UWORD Len)
{
  printf("%s: ",Prompt);
  fgets(Str, Len, stdin);
  Str[strlen(Str) - 1] = 0;
}

